package miss.bisolver;

import  juice.solver.*;
import    jam.solver.*;
import    jam.vector.*;
import    jam.operator.*;

/**
 * Conjugate Gradient Solver, plus a non-linear correction.<BR>
 *
 * @author    Joel Schroeder
 */
public class CGSolverNL extends IterativeSlv implements LinearSolver {
  protected Operator O;
  protected Vector   res;        //residual 
  protected float    rtrNew;     //norm2 of the residual
  protected int      reset = 50; //after each reset iterations the exact 

  public CGSolverNL(Operator O) {
    this(new IteratorSimple(), O); 
  }

  public CGSolverNL(Iterator iterator, Operator O) {
    this.iterator = iterator;
    this.iterator.setSolver(this); 
    this.O = O;
  }

  /**
   * after each reset-th iteration I compute the exact residual to 
   * remove accumulated floating point error (default is 50).
   */
  public void setReset(int reset) {
    this.reset = reset;
  }

  public Vector getResidual() {
    return res;
  }
  public float  getNormResidual() {
    return rtrNew;
  }

  public void solve(Operator A, Vector b, Vector x ) { 
    if (! (A instanceof LinearOperator)) {
      System.err.println("Error in CGSolver.solve(): " +
			 "operator A must be a LinearOperator");
    }

    this.A = A;
    this.b = b;
    this.x = x;

    res = A.getRange().newMember();
    A.residual(x,b,res);
    res.neg();
    O.addImage(res, res);
    Vector p = (Vector) res.copy();
    Vector v = A.getRange ().newMember();

    rtrNew      = res.norm2();

    iterator.start();

    while(iterator.hasMoreIterations()) {
      doOneStep(v, p);
      iterator.nextIteration(); 
      iter++;
    }
    iterator.finish();
  }

  protected void doOneStep(Vector v, Vector p) {
    A.image(p, v);

    // compute step length alpha
    float ptv   = p.dot( v);
    float alpha = rtrNew/ptv;
    if (ptv == 0f) System.err.println("Error in CGSolver: " +
				      "divide by ptv (=0).");
	
    // update x and r
	
    x.addScale(  alpha, p);
    //    if (((iter+1) % reset) == 0) {
    A.residual(x,b,res);
    res.neg();
    O.addImage(res, res);
    //    }
    //    else {
    //      res.addScale( -alpha, v);
    //    }
	
    // compute beta and the new search direction p
    
    float rtrOld = rtrNew;
    rtrNew       = res.norm2();
    float beta = rtrNew/rtrOld;
    if (rtrOld == 0f) System.err.println("Error in CGSolver: " +
				      "divide by rtrOld (=0).");
    p.addScale( beta, res, p);
  }

  
  // The following is to be added to the beginning of solve()
  /*
    if (!( A.getDomain()
      .equals(
      A.getRange ()) )) {
      System.err.println( "Error in conjugate gradient solver:\n" +
      "      Domain of A not equal to Range of A." );
      return;
    }
      if (!( A.getRange ().hasMember( b ) )) {
      System.err.println( "Error in conjugate gradient solver:" +
      "      b not in Range of A.");
      return;
    }
    if (!( A.getDomain().hasMember( x ) )) {
      System.err.println(  "Error in conjugate gradient solver:\n"
      + "      x not in Domain of A.\n" );
      return;
    }
    */
}
