package miss.misfilt;

import   sep.operator.SepOperator;
import   jam.operator.*;
import   jam.vector.*;
import   rsf.vector.*;
import   jam.JamObject;

/**
 * A <TT>SupportOp</TT> is an operator from an <TT>RsfSpace</TT> to itself
 * in which the output is zero whereever the input is zero and one wherever
 * the input is non-zero.<BR>
 *
 * <BR><!--#include file="./source.inc"--><BR>
 */
public class SupportOp extends SepOperator {
  protected int       dim;
  protected SupportOp complement = null;
  protected boolean   isCmp      = false;

  public     SupportOp(RsfSpace domainAndRange) {
    super(domainAndRange);
    dim = domainAndRange.getDimension();
  }

  /** copy constructor */
  protected  SupportOp(SupportOp src) {
    super(src);
    this.isCmp = src.isCmp;
    this.dim   = src.dim;
  }

  void createComplement() {
    try {
      this.complement            =  (SupportOp) this.clone();
    }
    catch (Exception e) { System.err.println(e); }
    this.complement.isCmp      = !this.isCmp;
    this.complement.complement =  this;
  }

  /**
   * The complement of a <TT>SupportOp</TT>, S, is the operator, C, which returns
   * the complement of the support of the <TT>Rsf</TT>.<BR>
   * Thus, whereas, S(x) is 0 where x is zero and 1 where x is non-zero,
   * C(x) is 1 where x is zero and 0 where x is non-zero.
   */
  public SupportOp getComplement() {
    if (complement == null) createComplement();
    return complement;
  }

  protected void apply(boolean add, Vector dom, Vector ran) {
    if (!add) ran.zero();
    switch (dim) {
    case 1: apply1D((Rsf1) dom, (Rsf1) ran); break;
    case 2: apply2D((Rsf2) dom, (Rsf2) ran); break;
    case 3: apply3D((Rsf3) dom, (Rsf3) ran); break;
    case 4: apply4D((Rsf4) dom, (Rsf4) ran); break;
    case 5: apply5D((Rsf5) dom, (Rsf5) ran); break;
    case 6: apply6D((Rsf6) dom, (Rsf6) ran); break;
    }
  }

  void apply1D(Rsf1 dom, Rsf1 ran) {
    float[] d = dom.d;
    float[] r = ran.d;
    int n = d.length;
    for (int i = 0; i < n; i++) {
      if (!isCmp) {
	if (d[i] != 0F)
	  r[i] += 1F;
      }
      else {
	if (d[i] == 0F)
	  r[i] += 1F;
      }
    }
  }

  void apply2D(Rsf2 dom, Rsf2 ran) {
    int n0 = ((RsfSpace) dom.getSpace()).getNSamples(0);
    int n1 = ((RsfSpace) dom.getSpace()).getNSamples(1);
    float[][] d = dom.d;
    float[][] r = ran.d;
    int i = 0;
    for   (int i1 = 0; i1 < n1; i1++)	       
      for (int i0 = 0; i0 < n0; i0++) {
	if (!isCmp) {
	  if (d[i1][i0] != 0F)
	    r[i1][i0] += 1F;
	}
	else {
	  if (d[i1][i0] == 0F)
	    r[i1][i0] += 1F;
	}
	i++;
      }
  }
  
  void apply3D(Rsf3 dom, Rsf3 ran) {
    int n0 = ((RsfSpace) dom.getSpace()).getNSamples(0);
    int n1 = ((RsfSpace) dom.getSpace()).getNSamples(1);
    int n2 = ((RsfSpace) dom.getSpace()).getNSamples(2);
    float[][][] d = dom.d;
    float[][][] r = ran.d;
    int i = 0;
    for     (int i2 = 0; i2 < n2; i2++)	       
      for   (int i1 = 0; i1 < n1; i1++)	       
	for (int i0 = 0; i0 < n0; i0++) {
	  if (!isCmp) {
	    if (d[i2][i1][i0] != 0F)
	      r[i2][i1][i0] += 1F;
	  }
	  else {
	    if (d[i2][i1][i0] == 0F)
	      r[i2][i1][i0] += 1F;
	  }
	  i++;
	}
  }
  
  void apply4D(Rsf4 dom, Rsf4 ran) {
    int n0 = ((RsfSpace) dom.getSpace()).getNSamples(0);
    int n1 = ((RsfSpace) dom.getSpace()).getNSamples(1);
    int n2 = ((RsfSpace) dom.getSpace()).getNSamples(2);
    int n3 = ((RsfSpace) dom.getSpace()).getNSamples(3);
    float[][][][] d = dom.d;
    float[][][][] r = ran.d;
    int i = 0;
    for	    (int i3 = 0; i3 < n3; i3++)	       
      for     (int i2 = 0; i2 < n2; i2++)	       
	for   (int i1 = 0; i1 < n1; i1++)	       
	  for (int i0 = 0; i0 < n0; i0++) {
	    if (!isCmp) {
	      if (d[i3][i2][i1][i0] != 0F)
		r[i3][i2][i1][i0] += 1F;
	    }
	    else {
	      if (d[i3][i2][i1][i0] == 0F)
		r[i3][i2][i1][i0] += 1F;
	    }
	    i++;
	  }
  }
  
  void apply5D(Rsf5 dom, Rsf5 ran) {
    int n0 = ((RsfSpace) dom.getSpace()).getNSamples(0);
    int n1 = ((RsfSpace) dom.getSpace()).getNSamples(1);
    int n2 = ((RsfSpace) dom.getSpace()).getNSamples(2);
    int n3 = ((RsfSpace) dom.getSpace()).getNSamples(3);
    int n4 = ((RsfSpace) dom.getSpace()).getNSamples(4);
    float[][][][][] d = dom.d;
    float[][][][][] r = ran.d;
    int i = 0;
    for   	(int i4 = 0; i4 < n4; i4++)	       
      for	(int i3 = 0; i3 < n3; i3++)	       
	for     (int i2 = 0; i2 < n2; i2++)	       
	  for   (int i1 = 0; i1 < n1; i1++)	       
	    for (int i0 = 0; i0 < n0; i0++) {
	      if (!isCmp) {
		if (d[i4][i3][i2][i1][i0] != 0F)
		  r[i4][i3][i2][i1][i0] += 1F;
	      }
	      else {
		if (d[i4][i3][i2][i1][i0] == 0F)
		  r[i4][i3][i2][i1][i0] += 1F;
	      }
	      i++;
	    }
  }
  
  void apply6D(Rsf6 dom, Rsf6 ran) {
    int n0 = ((RsfSpace) dom.getSpace()).getNSamples(0);
    int n1 = ((RsfSpace) dom.getSpace()).getNSamples(1);
    int n2 = ((RsfSpace) dom.getSpace()).getNSamples(2);
    int n3 = ((RsfSpace) dom.getSpace()).getNSamples(3);
    int n4 = ((RsfSpace) dom.getSpace()).getNSamples(4);
    int n5 = ((RsfSpace) dom.getSpace()).getNSamples(5);
    float[][][][][][] d = dom.d;
    float[][][][][][] r = ran.d;
    int i = 0;
    for       	    (int i5 = 0; i5 < n5; i5++)
      for   	    (int i4 = 0; i4 < n4; i4++)	       
	for	    (int i3 = 0; i3 < n3; i3++)	       
	  for       (int i2 = 0; i2 < n2; i2++)	       
	    for     (int i1 = 0; i1 < n1; i1++)	       
	      for   (int i0 = 0; i0 < n0; i0++) {
		if (!isCmp) {
		  if (d[i5][i4][i3][i2][i1][i0] != 0F)
		    r[i5][i4][i3][i2][i1][i0] += 1F;
		}
		else {
		  if (d[i5][i4][i3][i2][i1][i0] == 0F)
		    r[i5][i4][i3][i2][i1][i0] += 1F;
		}
		i++;
	      }
  }
  
  public JamObject copy() { return new SupportOp(this); }
}
