package  miss.misnonlinF;

import    jam.vector.*;
import    jam.operator.*;
import    rsfF.vector.*;
import    rsfF.solver.Mask;
import    sep.operator.*;
import  juice.operator.*;

/**
 * Creates a new binary operator from five pieces of information:
 * <UL>
 *     <LI> A bilinear operator B
 *     <LI> A vector in domain1 of B, k1
 *     <LI> A <TT>Mask</TT> which acts as a free mask in domain1, M1
 *     <LI> A vector in domain2 of B, k2
 *     <LI> A <TT>Mask</TT> which acts as a free mask in domain2, M2
 * </UL>
 * 
 * The operator created is A which is defined as: <BR>
 * <BR>
 * A(x1,x2) = B(k1 + M1(x1), k2 + M(x2))          <BR>
 * <BR>
 * This is not bilinear because of the k1 and k2 terms
 * 
 * <BR><!--#include file="./source.inc"--><BR>
 *
 * @author    Joel Schroeder
 */
public class BinaryMissingDataOp
extends SepBinaryOperator
implements hasHessian {
  SepBiLinearOperator B;
  Vector              k1, k2;
  Mask                M1, M2;
  
  public BinaryMissingDataOp(SepBiLinearOperator B,
			     Vector known1, Mask freeMask1,
			     Vector known2, Mask freeMask2) {
    super(B.getDomain1(), B.getDomain2(), B.getRange());
    this.B  = B;
    this.k1 = known1;
    this.k2 = known2;
    this.M1 = freeMask1;
    this.M2 = freeMask2;
  }

  public SepBiLinearOperator getSubOp () { return B ; }
  public Mask   	     getMask1 () { return M1; }
  public Mask   	     getMask2 () { return M2; }
  public Vector 	     getKnown1() { return k1; }
  public Vector 	     getKnown2() { return k2; }

  protected void apply(boolean add, Vector in1, Vector in2,
		       Vector  out) {
    Vector v1 = M1.image(in1);
    v1.add(k1);
    Vector v2 = M2.image(in2);
    v2.add(k2);

    if (add) {
      B.addImageBinary(v1, v2, out);
    }
    else {
      B.   imageBinary(v1, v2, out);
    }
  }

  public LinearOperator getGradientAt  (Vector x) {
    return getDerivativeAt(x).adjoint();
  }

  public hasAdjoint     getDerivativeAt(Vector x) {
    //    return new BMDOderiv(this, x);
    Vector at1 = ((ProductVector) x).getSubVector(0);
    Vector at2 = ((ProductVector) x).getSubVector(1);

    Vector v1 = M1.image(at1); v1.add(k1);
    Vector v2 = M2.image(at2); v2.add(k2);
    hasAdjoint L1 = new CompoundLinOpAdj(M1, (hasAdjoint) B.getOp2(v2));
    hasAdjoint L2 = new CompoundLinOpAdj(M2, (hasAdjoint) B.getOp1(v1));
    
    BlockLinearOperatorAdjoint
      deriv = new BlockLinOpAdj(B.getDomain(), 2,
				B.getRange (), 1);
    deriv.setOp(0,0, L1);
    deriv.setOp(0,1, L2);
    return deriv;
  }

  public BiLinearOperatorAdjoint getHessianAt(Vector x) {
    return new SepBiLinOpHessian(B);
  }

  public jam.JamObject copy() {
    System.err.println("BinaryMissingDataOp.copy() not implemented.\n");
    return null;
  }

  public boolean equals(Object obj2) {
    if (this == obj2) return true;
    if (! (obj2 instanceof BinaryMissingDataOp)) return false;
    BinaryMissingDataOp bop2 = (BinaryMissingDataOp) obj2;
    return (this.B .equals(bop2.B ) &&
	    this.k1.equals(bop2.k1) &&
	    this.k2.equals(bop2.k2) &&
	    this.M1.equals(bop2.M1) &&
	    this.M2.equals(bop2.M2));
  }

  public String  toString()   {
    return ("BinaryMissingDataOp (toString not implemented)");
  }
}
/*
final class BMDOderiv extends SepLinearOperator {
  BinaryMissingDataOp BMDO;
  Vector              v1;
  Vector              v2;

  public BMDOderiv(BinaryMissingDataOp B, Vector x) {
    super(B.getDomain(), B.getRange());
    BMDO = B;
    v1   = B.M1.image(((ProductVector) x).getSubVec(0));
    v2   = B.M2.image(((ProductVector) x).getSubVec(1));
    v1.add(B.k1);
    v2.add(B.k2);
  }

  void  apply(boolean add, Vector in, Vector out) {
    Vector in1 = ((ProductVector) in).getSubVec(0);
    Vector in2 = ((ProductVector) in).getSubVec(1);

    if (isFwd) {
      if (add)
	BMDO.B.addImageBinary(BMDO.M1.image(in1), v2, out);
      else
	BMDO.B.   imageBinary(BMDO.M1.image(in1), v2, out);
      BMDO.B.  addImageBinary(v1, BMDO.M2.image(in2), out);
    }
    else {
      if (add) {
	BMDO.M1.addImage(((hasAdjoint) BMDO.B.getOp2(v2)).adjoint().image(out),
			 in1);
	BMDO.M2.addImage(((hasAdjoint) BMDO.B.getOp1(v1)).adjoint().image(out),
			 in2);
      }
      else {
	BMDO.M1.   image(((hasAdjoint) BMDO.B.getOp2(v2)).adjoint().image(out),
			 in1);
	BMDO.M2.   image(((hasAdjoint) BMDO.B.getOp1(v1)).adjoint().image(out),
			 in2);
      }
    }
  }

  public jam.JamObject copy() {
    System.err.println("BinaryMissingDataOp.copy() not implemented.\n");
    return null;
  }

  public boolean equals(Object obj2) {
    if (this == obj2) return true;
    if (! (obj2 instanceof BinaryMissingDataOp)) return false;
    BinaryMissingDataOp bop2 = (BinaryMissingDataOp) obj2;
    return (this.B .equals(bop2.B ) &&
	    this.k1.equals(bop2.k1) &&
	    this.k2.equals(bop2.k2) &&
	    this.M1.equals(bop2.M1) &&
	    this.M2.equals(bop2.M2));
  }

  public String  toString()   {
    return ("BMDOderiv (toString not implemented)");
  }
}
*/
